#!/usr/bin/env python3

"""Print basic info on invoices."""

import asyncio
import json

from pytekukko.examples import example_argparser, example_client


async def run_example() -> None:
    """Run the example."""

    client, cookie_jar, cookie_jar_path = example_client(
        example_argparser(__doc__).parse_args()
    )

    data = []
    async with client.session:
        for invoice_header in await client.get_invoice_headers():
            data.append(
                {
                    "name": invoice_header.name,
                    "due_date": invoice_header.due_date.isoformat(),
                    "total": invoice_header.total,
                }
            )
        if not cookie_jar_path:
            await client.logout()

    print(json.dumps(data))

    if cookie_jar_path:
        cookie_jar.save(cookie_jar_path)


def main() -> None:
    """Run example in event loop."""
    asyncio.run(run_example())


if __name__ == "__main__":
    main()



"""Pytekukko tests."""

import datetime
import os
from typing import Any, Dict, TypeVar
from urllib.parse import parse_qs, quote_plus, urlparse, urlunparse

import pytest
from aiohttp import ClientSession

from pytekukko import Pytekukko
from pytekukko.examples import load_pytekukko_dotenv

T = TypeVar("T", bound=Dict[str, Any])  # pylint: disable=invalid-name

FAKE_CUSTOMER_NUMBER = "00-0000000-00"
FAKE_PASSWORD = "secret"  # noqa: S105
FAKE_POS = "1234"

QUERY_PARAMETER_FILTERS = [
    ("customerId", FAKE_CUSTOMER_NUMBER),
    ("customerNumber", FAKE_CUSTOMER_NUMBER),
    ("pos", FAKE_POS),
]


def before_record_response(response: T) -> T:
    """Scrub unwanted data before recording response."""

    response["headers"].pop("Set-Cookie", None)

    # As of vcrpy 4.1.1, query parameters filtered with filter_query_parameters
    # do not end up filtered in response["url"], so address them here.
    url_parts = urlparse(response["url"])
    new_query_parts = []
    query_params = parse_qs(url_parts.query)
    for key, values in query_params.items():
        for filter_key, filter_value in QUERY_PARAMETER_FILTERS:
            if key == filter_key:
                values = [filter_value]
        for value in values:
            new_query_parts.append(f"{quote_plus(key)}={quote_plus(value)}")
    new_url_parts = list(url_parts)
    new_url_parts[4] = "&".join(new_query_parts)
    response["url"] = urlunparse(new_url_parts)

    if response["url"].endswith("/login.do"):
        response["body"]["string"] = b"redacted"  # unused, bloats cassettes

    return response


@pytest.fixture(scope="module", autouse=True)
def load_dotenv() -> None:
    """Load our environment."""
    _ = load_pytekukko_dotenv()


@pytest.fixture(scope="module")
def vcr_config() -> Dict[str, Any]:
    """Get vcrpy configuration."""
    return {
        "before_record_response": before_record_response,
        "filter_headers": ["Cookie"],
        "filter_query_parameters": QUERY_PARAMETER_FILTERS,
        # NOTE: this should be uncommented when upgrading vcrpy to a fixed > 4.1.1:
        #   https://github.com/kevin1024/vcrpy/issues/398
        #   https://github.com/kevin1024/vcrpy/pull/582
        #   https://github.com/kevin1024/vcrpy/pull/581
        # "filter_post_data_parameters": ["j_username", "j_password"],
    }


@pytest.fixture(name="client", scope="function")
async def fixture_client() -> Pytekukko:
    """Get a client."""
    return Pytekukko(
        session=ClientSession(),
        customer_number=os.environ.get(
            "PYTEKUKKO_CUSTOMER_NUMBER", FAKE_CUSTOMER_NUMBER
        ),
        password=os.environ.get("PYTEKUKKO_PASSWORD", FAKE_PASSWORD),
    )


@pytest.mark.vcr
async def test_login(client: Pytekukko) -> None:
    """Test login."""
    async with client.session:
        result = await client.login()
    assert result


@pytest.mark.vcr
async def test_logout(client: Pytekukko) -> None:
    """Test logout."""
    # TODO(scop): would be better to test from a logged in session
    async with client.session:
        await client.logout()
    # No exception counts as success here


@pytest.mark.vcr
async def test_get_collection_schedule(client: Pytekukko) -> None:
    """Test getting collection schedule."""
    async with client.session:
        dates = await client.get_collection_schedule(
            what=int(os.environ.get("PYTEKUKKO_TEST_POS", FAKE_POS))
        )
    assert dates
    assert all(isinstance(date, datetime.date) for date in dates)


@pytest.mark.vcr
async def test_get_invoice_headers(client: Pytekukko) -> None:
    """Test getting invoice headers."""
    async with client.session:
        invoice_headers = await client.get_invoice_headers()
    assert invoice_headers
    assert all(invoice_header.raw_data for invoice_header in invoice_headers)
    assert all(invoice_header.name for invoice_header in invoice_headers)
    assert all(invoice_header.due_date for invoice_header in invoice_headers)
    assert all(invoice_header.total for invoice_header in invoice_headers)



#!/usr/bin/env python3

"""
Update Google Calendar with events for next collections.

To get the required service account file, enable the Google Calendar API
in the Google Cloud console, create service account credentials with
access to it, and create keys of type JSON.

Calendar id is typically the target Google Calendar account e-mail
address.
"""

import asyncio
import datetime
import logging
import sys
from typing import Any, Dict, Mapping, NamedTuple, cast

try:
    import zoneinfo  # type: ignore[import]
except ImportError:  # Python < 3.9
    from backports import zoneinfo  # type: ignore[import]

from google.oauth2 import service_account  # type: ignore[import]
from googleapiclient.discovery import build  # type: ignore[import]

from pytekukko.examples import arg_environ_default, example_argparser, example_client

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)


class CalendarData(NamedTuple):
    """Helper container for calendar data to update."""

    name: str
    date: datetime.date
    location: str


# https://developers.google.com/resources/api-libraries/documentation/calendar/v3/python/latest/calendar_v3.events.html
def update_google_calendar(  # pylint: disable=too-many-locals
    credentials: service_account.Credentials,
    calendar_id: str,
    data: Mapping[str, CalendarData],
) -> None:
    """Update the calendar."""
    service = build(
        "calendar",
        "v3",
        credentials=credentials,
        # https://github.com/googleapis/google-api-python-client/issues/299
        # https://github.com/googleapis/google-api-python-client/issues/325
        cache_discovery=False,
    )

    # Events with only date set are apparently treated as occurring at midnight,
    # start of day in calendar's timezone.
    calendar = (
        service.calendars()  # pylint: disable=no-member
        .get(calendarId=calendar_id)
        .execute()
    )
    if calendar.get("timeZone"):
        timezone: datetime.tzinfo = zoneinfo.ZoneInfo(calendar["timeZone"])
    else:
        timezone = datetime.timezone.utc
    since = datetime.datetime.now(timezone).replace(hour=0) - datetime.timedelta(
        hours=1
    )

    events = (
        service.events()  # pylint: disable=no-member
        .list(
            calendarId=calendar_id,
            orderBy="startTime",
            singleEvents=True,
            timeMin=since.isoformat(),
            privateExtendedProperty="pytekukko-managed=true",
        )
        .execute()
    )

    pos_events: Dict[str, Dict[str, Any]] = {}
    for event in events["items"]:
        # Store first event for each pos for update, delete rest
        pos = event["extendedProperties"]["private"].get("pytekukko-pos")
        if pos and pos in data and not pos_events.get(pos):
            pos_events[pos] = event
        else:
            service.events().delete(  # pylint: disable=no-member
                calendarId=calendar_id, eventId=event["id"]
            ).execute()

    for pos, pos_event_data in data.items():
        date = pos_event_data.date.isoformat()
        name = pos_event_data.name or ""
        description = f"{name} [pos={pos}]".strip()
        event_data = {
            "summary": "Jätekukko collection",
            "description": description,
            "location": pos_event_data.location or None,
            "start": {"date": date},
            "end": {"date": date},
            # Reminders are "for the authenticated user" per docs.
            # So for the service account, not the calendar owner :(
            # No way to set this for the "actual" calendar user from here,
            # at least while using a service account.
            "reminders": {"useDefault": False},
            "transparency": "transparent",
            "extendedProperties": {
                "private": {"pytekukko-managed": "true", "pytekukko-pos": pos}
            },
            "source": {
                "url": "https://tilasto.jatekukko.fi/indexservice2.jsp",
                "title": "Omakukko",
            },
            "creator": {
                "displayName": "Pytekukko",
            },
        }

        event = pos_events.get(pos)

        method = None
        log_level = logging.INFO
        if event:
            for key, value in event_data.items():
                if event.get(key) != value:
                    action = "updated"
                    method = service.events().patch(  # pylint: disable=no-member
                        calendarId=calendar_id, eventId=event["id"], body=event_data
                    )
                    break
            else:
                action = "unchanged"
                log_level = logging.DEBUG
        else:
            action = "created"
            method = service.events().insert(  # pylint: disable=no-member
                calendarId=calendar_id, body=event_data
            )
        if method:
            event = method.execute()
        LOGGER.log(log_level, "Event %s: %s", action, event.get("htmlLink"))


async def run_example() -> None:
    """Run the example."""

    argparser = example_argparser(__doc__)
    argparser.add_argument(
        "--google-calendar-id",
        type=str,
        **arg_environ_default("PYTEKUKKO_GOOGLE_CALENDAR_ID"),  # type: ignore[arg-type]
    )
    argparser.add_argument(
        "--google-service-account-file",
        type=str,
        **arg_environ_default(  # type: ignore[arg-type]
            "PYTEKUKKO_GOOGLE_SERVICE_ACCOUNT_FILE", fallback="service_account.json"
        ),
    )
    args = argparser.parse_args()

    if not args.google_calendar_id:
        print("Google calendar id required", file=sys.stderr)
        sys.exit(2)

    client, cookie_jar, cookie_jar_path = example_client(args)

    credentials = service_account.Credentials.from_service_account_file(
        args.google_service_account_file
    )

    async with client.session:
        services = await client.get_services()
        if not cookie_jar_path:
            await client.logout()

    if cookie_jar_path:
        cookie_jar.save(cookie_jar_path)

    data = {}
    for service in (x for x in services if x.next_collection):
        data[str(service.pos)] = CalendarData(
            name=service.name,
            date=cast(datetime.date, service.next_collection),
            location=", ".join(
                x
                for x in (
                    service.raw_data.get("owner", {}).get("katu"),
                    service.raw_data.get("owner", {}).get("posti"),
                )
                if x
            ),
        )

    await asyncio.get_event_loop().run_in_executor(
        None, update_google_calendar, credentials, args.google_calendar_id, data
    )


def main() -> None:
    """Run example in event loop."""
    asyncio.run(run_example())


if __name__ == "__main__":
    main()



"""Pytekukko examples."""

import os
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Dict, Optional, Tuple

from aiohttp import ClientSession, CookieJar
from dotenv import find_dotenv, load_dotenv

from pytekukko import Pytekukko


def arg_environ_default(
    key: str, optional: bool = False, fallback: Optional[str] = None
) -> Dict[str, Optional[str]]:
    """Get kwargs for environment backed argument addition."""
    help_text = f"default: ${key} from environment"
    if optional:
        help_text += " (optional)"
    if fallback:
        help_text += f", or {fallback}"
    return {
        "default": os.environ.get(key, fallback),
        "help": help_text,
    }


def load_pytekukko_dotenv() -> bool:
    """Load our .env."""
    return load_dotenv(os.environ.get("PYTEKUKKO_DOTENV", find_dotenv(usecwd=True)))


def example_argparser(description: str) -> ArgumentParser:
    """Set up example argument parser."""
    _ = load_pytekukko_dotenv()

    argparser = ArgumentParser(
        description=description,
        epilog=(
            "Environment variable defaults are loaded from the path set in "
            "$PYTEKUKKO_DOTENV in environment, falling back to .env if not set. "
            "If the path is relative, it is searched in directories starting from the "
            "current directory, walking towards the file system root."
        ),
    )
    argparser.add_argument(
        "--customer-number",
        type=str,
        **arg_environ_default("PYTEKUKKO_CUSTOMER_NUMBER"),  # type: ignore[arg-type]
    )
    argparser.add_argument(
        "--password",
        type=str,
        **arg_environ_default("PYTEKUKKO_PASSWORD"),  # type: ignore[arg-type]
    )
    argparser.add_argument(
        "--cookie-jar-file",
        type=str,
        **arg_environ_default(  # type: ignore[arg-type]
            "PYTEKUKKO_COOKIE_JAR_FILE", optional=True
        ),
    )

    return argparser


def example_client(args: Namespace) -> Tuple[Pytekukko, CookieJar, Optional[Path]]:
    """Set up example client."""

    if not args.customer_number:
        print("customer number required", file=sys.stderr)
        sys.exit(2)
    if not args.password:
        print("password required", file=sys.stderr)
        sys.exit(2)
    cookie_jar = CookieJar()
    cookie_jar_path = None
    if args.cookie_jar_file:
        cookie_jar_path = Path(args.cookie_jar_file)
        if cookie_jar_path.exists():
            cookie_jar.load(cookie_jar_path)

    client = Pytekukko(
        session=ClientSession(cookie_jar=cookie_jar),
        customer_number=args.customer_number,
        password=args.password,
    )

    return client, cookie_jar, cookie_jar_path



#!/usr/bin/env python3

"""Print collection schedules."""

import asyncio
import json

from pytekukko.examples import example_argparser, example_client


async def run_example() -> None:
    """Run the example."""

    client, cookie_jar, cookie_jar_path = example_client(
        example_argparser(__doc__).parse_args()
    )

    data = []
    async with client.session:
        services = await client.get_services()
        for service in services:
            schedule = await client.get_collection_schedule(service)
            if schedule:
                data.append(
                    {
                        "name": service.name,
                        "collection_schedule": [x.isoformat() for x in schedule],
                    }
                )
        if not cookie_jar_path:
            await client.logout()

    print(json.dumps(data))

    if cookie_jar_path:
        cookie_jar.save(cookie_jar_path)


def main() -> None:
    """Run example in event loop."""
    asyncio.run(run_example())


if __name__ == "__main__":
    main()



"""Jätekukko Omakukko client."""

from datetime import date, datetime
from http import HTTPStatus
from typing import Any, Dict, List, Union, cast
from urllib.parse import urljoin

from aiohttp import ClientResponse, ClientResponseError, ClientSession

from .models import CustomerData, InvoiceHeader, Service

__version__ = "0.11.0"
DEFAULT_BASE_URL = "https://tilasto.jatekukko.fi/jatekukko/"


class Pytekukko:
    """Client for accessing Jätekukko Omakukko services."""

    def __init__(
        self,
        session: ClientSession,
        customer_number: str,
        password: str,
        base_url: str = DEFAULT_BASE_URL,
    ):
        """Set up client."""
        self.session = session
        self.customer_number = customer_number
        self.password = password
        self.base_url = base_url

    async def get_customer_data(self) -> Dict[str, List[CustomerData]]:
        """Get customer data."""
        url = urljoin(self.base_url, "secure/get_customer_datas.do")

        response_data = await self._request_with_retry(method="GET", url=url)

        return {
            customer_number: [CustomerData(raw_data=a_data) for a_data in data]
            for customer_number, data in _unmarshal(response_data).items()
        }

    async def get_services(self) -> List[Service]:
        """Get services."""
        url = urljoin(self.base_url, "secure/get_services_by_customer_numbers.do")
        params = {"customerNumbers[]": self.customer_number}

        response_data = await self._request_with_retry(
            method="GET", url=url, params=params
        )
        assert isinstance(response_data, (list, tuple))

        return [Service(raw_data=_unmarshal(service)) for service in response_data]

    async def get_collection_schedule(self, what: Union[Service, int]) -> List[date]:
        """
        Get collection schedule for a service.

        :param what: the service or a "pos" value of one to get schedule for
        """
        url = urljoin(self.base_url, "get_collection_schedule.do")
        pos = what.pos if isinstance(what, Service) else what
        params = {"customerNumber": self.customer_number, "pos": pos}

        response_data = await self._request_with_retry(
            method="GET", url=url, params=params
        )

        return cast(List[date], _unmarshal(response_data))

    async def get_invoice_headers(self) -> List[InvoiceHeader]:
        """Get headers of available invoices."""
        url = urljoin(self.base_url, "secure/get_invoice_headers_for_customer.do")
        params = {
            "customerId": self.customer_number,  # yep, customerId, not *Number here
        }

        response_data = await self._request_with_retry(
            method="GET", url=url, params=params
        )
        assert isinstance(response_data, (list, tuple))

        return [
            InvoiceHeader(raw_data=_unmarshal(invoice_header))
            for invoice_header in response_data
        ]

    async def login(self) -> Dict[str, str]:
        """Log in."""
        url = urljoin(self.base_url, "j_acegi_security_check")
        headers = (("X-Requested-With", "XMLHttpRequest"),)
        params = {"target": "2"}
        data = {"j_username": self.customer_number, "j_password": self.password}

        async with self.session.post(
            url, headers=headers, params=params, data=data, raise_for_status=True
        ) as response:
            # TODO(scop): Check we got {"response":"OK"}?
            return cast(Dict[str, str], await response.json())

    async def logout(self) -> None:
        """Log out the current session."""
        url = urljoin(self.base_url, "j_acegi_logout_elcustrap")

        async with self.session.get(url, raise_for_status=True) as response:
            await _drain(response)

    async def _request_with_retry(self, **request_kwargs: Any) -> Any:
        """
        Do a request, with automatic login and retry if session is logged out.

        :param raise_for_first_status: whether first unsuccessful status should raise;
            False allows for handling special cases that give errors instead of
            redirecting to login page
        :param request_kwargs: kwargs to pass to self.session.request
        """
        try:
            async with self.session.request(
                **request_kwargs, raise_for_status=True
            ) as response:
                if response.history and response.url.path.endswith("/login.do"):
                    await _drain(response)
                else:
                    return await response.json()
        except ClientResponseError as ex:
            if not (
                ex.status in (HTTPStatus.BAD_REQUEST, HTTPStatus.INTERNAL_SERVER_ERROR)
                and "get_collection_schedule" in ex.request_info.url.path
            ):
                raise

        _ = await self.login()
        async with self.session.request(
            **request_kwargs, raise_for_status=True
        ) as response:
            return await response.json()


def _unmarshal(data: Any) -> Any:
    """
    Unmarshal items in parsed JSON to more specific objects.

    :param data: parsed JSON data
    :return: unmarshalled data
    """
    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = _unmarshal(value)
    elif isinstance(data, list):
        for i, value in enumerate(data):
            data[i] = _unmarshal(value)
    elif isinstance(data, str):
        try:
            data = datetime.strptime(data, "%Y-%m-%d").date()
        except ValueError:
            try:
                data = datetime.strptime(data, "%H:%M").time()
            except ValueError:
                pass
    return data


async def _drain(response: ClientResponse) -> None:
    """
    Consume and discard response.

    Useful for keeping the connection alive without caring about response content.
    """
    async for _ in response.content.iter_chunked(1024):
        pass



"""Pytekukko model objects."""

from dataclasses import dataclass
from datetime import date
from typing import Any, Dict, Optional, cast


@dataclass
class Service:
    """
    Service encapsulates information about parts of a customer relationship.

    Examples of the kinds of services there are include collections of different kinds
    of waste containers, and yearly base prices for houses.

    Some frequently used service attributes are available as individual properties,
    and all data retrieved from the service is available in the ``raw_data`` dict.
    """

    raw_data: Dict[str, Any]

    @property
    def name(self) -> str:
        """Get service name."""
        return cast(str, self.raw_data["ASTNimi"])

    @property
    def pos(self) -> int:
        """Get "pos" value."""
        return cast(int, self.raw_data["ASTPos"])

    @property
    def next_collection(self) -> Optional[date]:
        """
        Get next collection date.

        :returns: Next collection date, None if not applicable for the service.
        """
        return self.raw_data.get("ASTSeurTyhj")


@dataclass
class CustomerData:
    """
    CustomerData encapsulates customer information.

    Some frequently used service attributes are available as individual properties,
    and all data retrieved from the service is available in the ``raw_data`` dict.
    """

    raw_data: Dict[str, Any]

    @property
    def customer_number(self) -> str:
        """Get customer number."""
        return cast(str, self.raw_data["asiakasnro"])

    @property
    def name(self) -> str:
        """Get customer name."""
        return cast(str, self.raw_data["nimi"])


@dataclass
class InvoiceHeader:
    """
    InvoiceHeader encapsulates basic information of an invoice.

    Some frequently used service attributes are available as individual properties,
    and all data retrieved from the service is available in the ``raw_data`` dict.
    """

    raw_data: Dict[str, Any]

    @property
    def name(self) -> str:
        """Get customer number."""
        return cast(str, self.raw_data["name"])

    @property
    def due_date(self) -> date:
        """Get due date."""
        return cast(date, self.raw_data["dueDate"])

    @property
    def total(self) -> float:
        """Get total amount."""
        return cast(float, self.raw_data["total"])



#!/usr/bin/env python3

"""Pytekukko setup."""

from setuptools import setup  # type: ignore[import]


def get_version() -> str:
    """Extract version number."""
    filename = "src/pytekukko/__init__.py"
    with open(filename, encoding="utf-8") as file_:
        for line in file_:
            if line.startswith("__version__"):
                return line.split("=")[-1].strip("\"'\r\n ")
    raise Exception(f"No __version__ in {filename}!")


if __name__ == "__main__":
    setup(version=get_version())
